"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy

def mask_from_model():
    pass

def half2full(map_h1, map_h2):
    assert map_h1.shape == map_h2.shape
    assert map_h1.unit_cell == map_h2.unit_cell
    tmp = (numpy.array(map_h1)+numpy.array(map_h2))/2.
    gr = gemmi.FloatGrid(tmp, map_h1.unit_cell, map_h1.spacegroup)
    return gr
# half2full()

def write_ccp4_map(filename, array, cell=None, sg=None):
    if type(array) == numpy.ndarray:
        # TODO check dtype
        if sg is None: sg = gemmi.SpaceGroup(1)
        grid = gemmi.FloatGrid(array, cell, sg)
    else:
        # TODO check type
        grid = array
        
    ccp4 = gemmi.Ccp4Map()
    ccp4.grid = grid
    ccp4.update_ccp4_header(2, True) # float, update stats
    ccp4.write_ccp4_map(filename)
# write_ccp4_map()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
from servalcat.utils import logger
from servalcat import utils
import argparse

var_cmpl = lambda x: numpy.var(x.real)+numpy.var(x.imag)
var_cmpl = lambda x: numpy.var(x)

def add_arguments(parser):
    parser.description = 'Fo-Fc map calculation based on model and data errors'
    parser.add_argument("--halfmaps", required=True, nargs=2)
    #parser.add_argument('--mapref', help='Reference map file')
    parser.add_argument('--model', required=True,
                        help='Input atomic model file (PDB or mmCIF/PDBx')
    parser.add_argument("-d", '--resolution', type=float, required=True)
    parser.add_argument('-m', '--mask', help="mask file")
    parser.add_argument('-r', '--mask_radius', type=float, help="mask radius")
    parser.add_argument("-B", type=float, help="Estimated blurring.")
    parser.add_argument("--normalized_map", action='store_true',
                        help="Write normalized map in the masked region")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--output_prefix', default="diffmap",
                        help='output file name prefix')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def read_maps(halfmaps, d_min, mask=None):
    assert len(halfmaps) == 2
    map1 = gemmi.read_ccp4_map(halfmaps[0])
    map2 = gemmi.read_ccp4_map(halfmaps[1])
    logger.write("SG1= {}".format(map1.grid.spacegroup.hm))
    logger.write("SG2= {}".format(map1.grid.spacegroup.hm))
    
    if mask is not None:
        map1.grid = gemmi.FloatGrid(numpy.array(map1.grid)*mask,
                                    map1.grid.unit_cell, map1.grid.spacegroup)
        map2.grid = gemmi.FloatGrid(numpy.array(map2.grid)*mask,
                                    map2.grid.unit_cell, map2.grid.spacegroup)

    asu1 = gemmi.transform_map_to_f_phi(map1.grid).prepare_asu_data(dmin=d_min)
    asu2 = gemmi.transform_map_to_f_phi(map2.grid).prepare_asu_data(dmin=d_min)

    return asu1, asu2
# read_maps()
    
def calc_noise_var(asu1, asu2, fc_asu):
    df = utils.hkl.df_from_asu_data(asu1, "F_map1")
    hkldata = utils.hkl.HklData(asu1.unit_cell, asu1.spacegroup, False, df)
    hkldata.merge_asu_data(asu2, "F_map2")
    hkldata.merge_asu_data(fc_asu, "FC")
    hkldata.df["FP"] = (hkldata.df.F_map1 + hkldata.df.F_map2)/2.
    #logger.write(hkldata.df)
    hkldata.setup_relion_binning()
    #logger.write(hkldata.df[hkldata.df.bin==1])
    # Scale
    #iniscale = utils.scaling.InitialScaler(fo_asu, fc_asu, aniso=True)
    #iniscale.run()
    #scale_for_fo = iniscale.get_scales()
    #fo_asu.value_array[:] *= scale_for_fo
    #asu1.value_array[:] *= scale_for_fo
    #asu2.value_array[:] *= scale_for_fo

    #logger.write("half_l=", gemmi.transform_map_to_f_phi(map1.grid).half_l)
    s_array = 1./hkldata.d_spacings()
    hkldata.binned_df["var_noise"] = 0.
    hkldata.binned_df["FSCfull"] = 0.
    #logger.write(hkldata.df)


    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        # scale
        #fo = numpy.array(hkldata.df.FP[sel])
        #fc = numpy.array(hkldata.df.FC[sel])
        fo = hkldata.df.FP[sel]
        fc = hkldata.df.FC[sel]
        scale = 1. #numpy.sqrt(var_cmpl(fc)/var_cmpl(fo))
        #hkldata.df.loc[sel, "FP"] *= scale
        #hkldata.df.loc[sel, "F_map1"] *= scale
        #hkldata.df.loc[sel, "F_map2"] *= scale
        
        sel1 = numpy.array(hkldata.df.F_map1[sel])
        sel2 = numpy.array(hkldata.df.F_map2[sel])

        fsc = numpy.real(numpy.corrcoef(sel1, sel2)[1,0])
        varn = var_cmpl(sel1-sel2)/4
        logger.write("{:3d} {:7d} {:7.3f} {:7.3f} {:.4f} {:e} {}".format(i_bin, sel1.size, bin_d_max, bin_d_min,
                                                                         fsc, varn, scale))
        hkldata.binned_df.loc[i_bin, "var_noise"] = varn
        hkldata.binned_df.loc[i_bin, "FSCfull"] = 2*fsc/(1+fsc)

    return hkldata
# calc_noise_var()

def calc_D_and_S(hkldata, output_prefix):#fo_asu, fc_asu, varn, bins, bin_idxes):
    hkldata.binned_df["D"] = 0.
    hkldata.binned_df["S"] = 0.

    ofs = open("{}_Fstats.dat".format(output_prefix), "w")
    ofs.write("bin       n   d_max   d_min         Fo         Fc FSC.model FSC.full      D          S          N\n")
    tmpl = "{:3d} {:7d} {:7.3f} {:7.3f} {:.4e} {:.4e} {: .4f}   {: .4f} {: .4e} {:.4e} {:.4e}\n"
    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        Fo = numpy.array(hkldata.df.FP[sel])
        Fc = numpy.array(hkldata.df.FC[sel])
        D = hkldata.binned_df.D
        S = hkldata.binned_df.S
        varn = hkldata.binned_df.var_noise[i_bin]
        fsc = numpy.real(numpy.corrcoef(Fo, Fc)[1,0])
        fsc_full = hkldata.binned_df.FSCfull
        D.loc[i_bin] = numpy.sum(numpy.real(Fo * numpy.conj(Fc)))/numpy.sum(numpy.abs(Fc)**2)
        #D.loc[i_bin] = numpy.sum(numpy.abs(Fo) * numpy.abs(Fc))/numpy.sum(numpy.abs(Fc)**2)
        S.loc[i_bin] = max(0, numpy.average(numpy.abs(Fo-D[i_bin]*Fc)**2)-varn)
        ofs.write(tmpl.format(i_bin, Fo.size, bin_d_max, bin_d_min,
                              numpy.average(numpy.abs(Fo)),
                              numpy.average(numpy.abs(Fc)),
                              fsc, fsc_full[i_bin], D[i_bin], S[i_bin], varn))
# calc_D_and_S()

def calc_maps(hkldata, B=None):#fo_asu, fc_asu, D, S, varn, bins, bin_idxes):
    labs = ["DELFWT", "FWT", "DELFWT_noscale", "FWT_noscale"]
    if B is not None: labs.extend(["DELFWT_b0", "FWT_b0"])
    for l in labs: hkldata.df[l] = 0.j
    logger.write("Calculating maps..")
    logger.write("NOTE: do not use FSC_local, but FSC")
    for i_bin, bin_d_max, bin_d_min in hkldata.bin_and_limits():
        sel = i_bin == hkldata.df.bin
        Fo = hkldata.df.FP[sel]
        Fc = hkldata.df.FC[sel]
        D = hkldata.binned_df.D[i_bin]
        S = hkldata.binned_df.S[i_bin]
        FSCfull = hkldata.binned_df.FSCfull[i_bin]
        varn = hkldata.binned_df.var_noise[i_bin]
        FSCfull = min(1e-6, FSCfull)

        delfwt = (Fo-D*Fc)*S/(S+varn)
        fwt = (Fo*S+varn*D*Fc)/(S+varn)

        hkldata.df.loc[sel, "DELFWT_noscale"] = delfwt
        hkldata.df.loc[sel, "FWT_noscale"] = fwt

        #sig_fo = numpy.sqrt(var_cmpl(Fo))
        sig_fo = numpy.std(numpy.array(Fo))
        n_fo = sig_fo * numpy.sqrt(FSCfull)
        #n_fofc = numpy.sqrt(var_cmpl(Fo-D*Fc))

        lab_suf = "" if B is None else "_b0"
        hkldata.df.loc[sel, "DELFWT"+lab_suf] = delfwt/n_fo
        hkldata.df.loc[sel, "FWT"+lab_suf] = fwt/n_fo

        if B is not None:
            s2 = 1./hkldata.d_spacings()[sel]**2
            k = numpy.exp(-B*s2/4.)
            k2 = numpy.exp(-B*s2/2.)
            fsc_l = k2*FSCfull/(1+(k2-1)*FSCfull)
            #n_fo = sig_fo * numpy.sqrt(fsc_l) * k
            S_l = S * k2
            
            delfwt = (Fo-D*Fc)*S*k/(S_l+varn)/n_fo # S_l/k = S*k
            fwt = (Fo*S_l+varn*D*Fc)/(S_l+varn)/n_fo/k
            logger.write("{:4d} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e} {:.4e}".format(i_bin,
                                                                           numpy.average(n_fo),
                                                                           numpy.average(sig_fo),
                                                                           numpy.average(fsc_l),
                                                                           numpy.average(k),
                                                                           numpy.average(abs(fwt)),
                                                                           numpy.average(abs(delfwt))))
            hkldata.df.loc[sel, "DELFWT"] = delfwt
            hkldata.df.loc[sel, "FWT"] = fwt
    return labs
# calc_maps()

def dump_to_mtz(hkldata, map_labs, mtz_out):
    map_labs = map_labs + ["FP", "FC"]
    data = numpy.empty((len(hkldata.df.index), len(map_labs)*2+3))
    data[:,:3] = hkldata.df[["H","K","L"]]
    for i, lab in enumerate(map_labs):
        data[:,3+i*2] = numpy.abs(hkldata.df[lab])
        data[:,3+i*2+1] = numpy.angle(hkldata.df[lab], deg=True)
        
    mtz = gemmi.Mtz()
    mtz.spacegroup = hkldata.sg
    mtz.cell = hkldata.cell
    mtz.add_dataset('HKL_base')
    for label in ['H', 'K', 'L']: mtz.add_column(label, 'H')

    for lab in map_labs:
        mtz.add_column(lab, "F")
        mtz.add_column(("PH"+lab).replace("FWT", "WT"), "P")
    
    #mtz.add_column("FWT", "F")
    #mtz.add_column("PHWT", "P")
    #mtz.add_column("DELFWT", "F")
    #mtz.add_column("PHDELWT", "P")
    #mtz.add_column("FP", "F")
    #mtz.add_column("PHFP", "P")
    #mtz.add_column("FC", "F")
    #mtz.add_column("PHFC", "P")
    mtz.set_data(data)
    mtz.write_to_file(mtz_out)
# dump_to_mtz()
        
def main(args):
    st = gemmi.read_structure(args.model)
    st.expand_ncs(gemmi.HowToNameCopiedChain.Short)

    g = gemmi.read_ccp4_map(args.halfmaps[0]).grid
    st.spacegroup_hm = "P1"
    st.cell = g.unit_cell

    if st[0].count_hydrogen_sites() > 0:
        monlib = utils.model.load_monomer_library(st[0].get_all_residue_names(),
                                                  monomer_dir=args.monlib)
    else:
        monlib = None

    if args.mask:
        mask = numpy.array(gemmi.read_ccp4_map(args.mask).grid)
    elif args.mask_radius:
        tmp = gemmi.read_ccp4_map(args.halfmaps[0]).grid
        mask = gemmi.FloatGrid(tmp.nu, tmp.nv, tmp.nw)
        mask.set_unit_cell(tmp.unit_cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
        mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)
        ccp4 = gemmi.Ccp4Map()
        ccp4.grid = mask
        ccp4.update_ccp4_header(2, True) # float, update stats
        ccp4.write_ccp4_map("mask_from_model.ccp4")
        mask = numpy.array(mask)
    else:
        mask = None
        if args.normalized_map:
            logger.write("WARNING: Mask is not available. --normalized-map will have no effect.")
        
    fc_asu = utils.model.calc_fc_em(st, args.resolution, r_cut=1e-7, monlib=monlib)
    asu1, asu2 = read_maps(args.halfmaps, args.resolution, mask)
    hkldata = calc_noise_var(asu1, asu2, fc_asu)
    #dump_to_mtz(fo_asu, "Fo.mtz")

    if args.B is not None:
        Bave = numpy.average(utils.model.all_B(st))
        logger.write("Using user-specified B: {}".format(args.B))
        logger.write("    Average B of model= {:.2f}".format(Bave))
        B = args.B - Bave
        logger.write("    Relative B for map= {:.2f}".format(B))
    else:
        B = None
        
    calc_D_and_S(hkldata, args.output_prefix)#fo_asu, fc_asu, varn_bin, bins, bin_idxes)
    map_labs = calc_maps(hkldata, B=B)#fo_asu, fc_asu, D, S, varn_bin, bins, bin_idxes)
    dump_to_mtz(hkldata, map_labs, "{}.mtz".format(args.output_prefix))

    if args.normalized_map and mask is not None:
        logger.write("Normalized Fo-Fc map requested.")
        delfwt_map = hkldata.fft_map("DELFWT", grid_size=mask.shape)
        cutoff = 1. - 1e-6
        masked = numpy.array(delfwt_map)[mask>cutoff]
        logger.write("   Whole volume: {} voxels".format(delfwt_map.point_count))
        logger.write("  Masked volume: {} voxels (>{})".format(masked.size, cutoff))
        global_mean = numpy.average(delfwt_map)
        global_std = numpy.std(delfwt_map)
        logger.write("    Global mean: {}".format(global_mean))
        logger.write("     Global std: {}".format(global_std))
        masked_mean = numpy.average(masked)
        masked_std = numpy.std(masked)
        logger.write("    Masked mean: {}".format(masked_mean))
        logger.write("     Masked std: {}".format(masked_std))
        #logger.write(" If you want to scale manually: {}".format())
        scaled = (delfwt_map - masked_mean)/masked_std
        filename = "{}_normalized_fofc.mrc".format(args.output_prefix)
        logger.write("  Writing {}".format(filename))
        utils.maps.write_ccp4_map(filename, scaled, cell=st.cell)

# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""

__version__ = '0.0.1'



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import os
import gemmi
import numpy
import argparse
import json
from servalcat.utils import logger
from servalcat import utils

def add_arguments(parser):
    parser.description = 'Shift maps and models into new small box'
    parser.add_argument('--maps',
                        required=True,
                        nargs="+",
                        help='Input map file(s)')
    parser.add_argument('--mask',
                        required=False,
                        help='Mask file')    
    parser.add_argument('--model',
                        required=False,
                        nargs="+",
                        help='Input atomic model file')
    parser.add_argument('--output_prefix',
                        required=False,
                        default="newbox",
                        help='outpu file name prefix')
    parser.add_argument('--mask_radius',
                        required=False,
                        type=float,
                        help='Define new box using model')
    parser.add_argument('--noncubic',
                        action='store_true')
    parser.add_argument('--force_cell',
                        required=False,
                        type=float,
                        nargs=6,
                        help='Force cell')
    parser.add_argument('--disable_cell_check',
                        action='store_true')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def check_maps(map_files, disable_cell_check=False):
    logger.write("Input map files:")
    params = []
    for f in map_files:
        m = gemmi.read_ccp4_map(f)
        g = m.grid
        logger.write(" {}".format(f))
        logger.write("     Cell: {}".format(g.unit_cell.parameters))
        logger.write("     Grid: {}".format(g.shape))
        logger.write("    Start: {}".format([m.header_i32(x) for x in (5,6,7)]))
        logger.write("  Spacing: {}".format(g.spacing))
        logger.write(" Space group: {}".format(g.spacegroup.hm))
        params.append((g.unit_cell.parameters, g.shape, g.spacing))

    shapes = set([x[1] for x in params])
    if len(shapes) > 1:
        raise RuntimeError("Error: different grid size included")

    cells = set([x[0] for x in params])
    if len(cells) > 1:
        if disable_cell_check:
            logger.write("WARNING: Cells are different. Using the first one.")
        else:
            raise RuntimeError("Error: different cell parameters included")

    return params[0]
        
def main(args):
    cell, grid_shape, spacing = check_maps(args.maps, args.disable_cell_check)
    if args.force_cell:
        cell = args.force_cell

    cell = gemmi.UnitCell(*cell)
    sts, cif_refs = [], []
    if args.model:
        for m in args.model:
            st, cif_ref = utils.fileio.read_structure_from_pdb_and_mmcif(m)
            st.spacegroup_hm = "P1"
            st.cell = cell
            sts.append(st)
            cif_refs.append(cif_ref)
        
    if args.mask:
        logger.write("Using mask to decide border: {}".format(args.mask))
        mask = gemmi.read_ccp4_map(args.mask).grid
        mask.set_unit_cell(cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
    elif args.model:
        logger.write("Using model to decide border: {}".format(args.model))
        mask = gemmi.FloatGrid(*grid_shape)
        mask.set_unit_cell(cell)
        mask.spacegroup = gemmi.SpaceGroup(1)
        for st in sts:
            mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)
    else:
        raise RuntimeError("Give mask or model")

    tmp = numpy.where(numpy.array(mask)>0)
    limits = [(min(x), max(x)) for x in tmp]
    # Option to keep cubic?
    spacing = numpy.array(spacing)
    p = args.mask_radius/spacing if args.mask_radius else 3/spacing
    p = p.astype(int)
    logger.write("Limits: {}".format(limits))
    logger.write("Padding: {}".format(p))
    new_shape = [0, 0, 0]
    min_points = [0, 0, 0]
    slices = [0, 0, 0]
    rads = [0, 0, 0]
    for i in range(3):
        ctr = (grid_shape[i]-1)/2 
        rad = max(ctr-(limits[i][0]-p[i]), (limits[i][1]+p[i])-ctr)
        logger.write("Rad{}= {}".format(i, rad))
        if rad < grid_shape[i]/2:
            slices[i] = slice(int(ctr-rad), int(ctr+rad)+1, None)
        else:
            slices[i] = slice(0, grid_shape[i], None)
            
        min_points[i] = slices[i].start

    logger.write("Slices: {}".format(slices))
    if not args.noncubic:
        min_s = min([slices[i].start for i in range(3)])
        max_s = max([slices[i].stop for i in range(3)])
        slices = [slice(min_s, max_s, None) for i in range(3)]
        min_points = [min_s for i in range(3)]
        logger.write("Cubic Slices: {}".format(slices))

    if args.model:
        shifts = -mask.point_to_position(slices[0].start, slices[1].start, slices[2].start)
        logger.write("Shift for model: {}".format(shifts))
        for st in sts:
            for mol in st:
                for cra in mol.all():
                    cra.atom.pos += shifts

    new_shape = [slices[0].stop-slices[0].start,
                 slices[1].stop-slices[1].start,
                 slices[2].stop-slices[2].start]
    tmp = mask.point_to_position(*new_shape)
    new_cell = gemmi.UnitCell(tmp[0], tmp[1], tmp[2], cell.alpha, cell.beta, cell.gamma)

    if args.model:
        for i, st in enumerate(sts):
            st.cell = new_cell
            if len(st.ncs) > 0:
                new_ops = utils.symmetry.apply_shift_for_ncsops(st.ncs, shifts)
                st.ncs.clear()
                st.ncs.extend(new_ops)
                logger.write(" Writing symmetry expanded model for shifted model")
                utils.symmetry.write_symmetry_expanded_model(st, "shifted_local_expanded",
                                                             pdb=True, cif=True)

            spext = utils.fileio.splitext(os.path.basename(args.model[i]))
            utils.fileio.write_model(st, file_name=spext[0]+"_cut"+spext[1], cif_ref=cif_refs[i])
    
    for f in args.maps:
        logger.write("Slicing {}".format(f))
        g = numpy.array(gemmi.read_ccp4_map(f).grid)
        newg = g[slices[0], slices[1], slices[2]]
        ccp4 = gemmi.Ccp4Map()
        ccp4.grid = gemmi.FloatGrid(newg, new_cell, mask.spacegroup)
        ccp4.update_ccp4_header(2, True) # float, update stats
        ccp4.write_ccp4_map(os.path.basename(utils.fileio.splitext(f)[0])+"_cut.mrc")
    
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import argparse
import json
from servalcat.utils import logger
from servalcat import utils

def add_arguments(parser):
    parser.description = 'Structure factor calculation for Refmac'
    parser.add_argument('--map',
                        required=False,
                        help='Input map file')
    parser.add_argument('--halfmaps',
                        required=False,
                        nargs=2,
                        help='Input half map files')
    parser.add_argument('--mapref',
                        required=False,
                        help='Reference map file')
    parser.add_argument('--mask',
                        required=False,
                        help='Mask file')
    parser.add_argument('--model',
                        required=False,
                        help='Input atomic model file')
    parser.add_argument('--output_masked_prefix',
                        required=False,
                        default="masked_fs",
                        help='output MTZ file name for masked F')
    parser.add_argument('--output_mtz_prefix',
                        required=False,
                        default="starting_map",
                        help='output MTZ file name for original F')
    parser.add_argument('--output_model_prefix',
                        required=False,
                        default="shifted_local",
                        help='output model file name')
    parser.add_argument('--mask_radius',
                        required=False,
                        type=float,
                        help='')
    parser.add_argument('--resolution',
                        required=True,
                        type=float,
                        help='')
    parser.add_argument('--shift',
                        action='store_true',
                        help='')
    parser.add_argument('--blur',
                        required=False,
                        nargs="+",
                        type=float,
                        help='Sharpening or blurring B')
    parser.add_argument('--relion_pg',
                        required=False,
                        help='RELION point group symbol for strict symmetry')
    parser.add_argument('--ignore_symmetry',
                        required=False,
                        help='Ignore symmetry information in the model file')
    parser.add_argument('--remove_multiple_models',
                        help='Keep 1st model only')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def write_map_mtz(asu_data, mtz_out, blurs=None):
    if not blurs: blurs = []
    if 0 not in blurs:
        blurs.insert(0, 0)
    
    mtz = gemmi.Mtz()
    mtz.spacegroup = asu_data.spacegroup
    mtz.cell = asu_data.unit_cell
    mtz.add_dataset('HKL_base')
    for label in ['H', 'K', 'L']: mtz.add_column(label, 'H')
    for b in blurs:
        if b == 0:
            mtz.add_column("Fout0", "F")
        else:
            mtz.add_column("FoutBlur_{:.2f}".format(b), "F")

    mtz.add_column("Pout0", "P")

    data = numpy.empty((len(asu_data), 3+len(blurs)+1))
    data[:,:3] = asu_data.miller_array
    s2 = asu_data.make_1_d2_array()
    
    for i, b in enumerate(blurs):
        if b == 0:
            data[:,3+i] = numpy.absolute(asu_data.value_array)
        else:
            tempfac = numpy.exp(-b*s2/4.)
            data[:,3+i] = numpy.absolute(asu_data.value_array * tempfac)
            
    data[:,-1] = numpy.angle(asu_data.value_array, deg=True)

    mtz.set_data(data)
    mtz.write_to_file(mtz_out)
# write_map_mtz()

def scale_maps(maps_in, map_ref, d_min):
    fs = []
    for m in maps_in+[map_ref]:
        asu = gemmi.transform_map_to_f_phi(m).prepare_asu_data(dmin=d_min)
        fs.append(asu)

    binner = utils.hkl.Binner(fs[0], style="relion")
    #fs, binner = fft_and_binning([map_ref]+maps_in, d_min)
    f_ref, f_inps = fs[0], fs[1:]
    d_array = f_ref.make_d_array()
    f_inps_scaled = [f_obs.copy() for f_obs in f_inps]
    logger.write(" Saving sf stats as f_stats.dat")
    ofs = open("f_stats.dat", "w")
    lab_fos = " ".join(["F_obs.{n} F_scaled.{n}".format(n=i+1) for i in range(len(f_inps))])
    ofs.write("bin count d_min F_ref {}\n".format(lab_fos))
    for i_bin, c_bin in zip(binner.bins, binner.bin_counts):
        sel = binner.bin_array == i_bin
        fr = f_ref.value_array[sel]
        s_fr = numpy.std(fr)
        d = min(d_array[sel])
        avg_strs = []
        for f_obs, f_obs_scaled in zip(f_inps, f_inps_scaled):
            fo = f_obs.value_array[sel]
            s_fo = numpy.std(fo)
            fo_scaled = fo*s_fr/s_fo
            f_obs_scaled.value_array[sel] = fo_scaled
            avg_strs.append("{:e} {:e}".format(numpy.average(numpy.abs(fo)), numpy.average(numpy.abs(fo_scaled))))

        ofs.write("{:3d} {:6d} {:7.3f} {:e} {}\n".format(i_bin, c_bin, d,
                                                         numpy.average(numpy.abs(fr)), " ".join(avg_strs)))
                                                                

    maps_scaled = [gemmi.transform_f_phi_grid_to_map(f.get_f_phi_on_grid(size=map_ref.shape))
                   for f in f_inps_scaled]
    return maps_scaled
# scale_maps()

def main(args):
    if args.mask and args.mask_radius:
        logger.write("You cannot give mask and mask_radius")
        return
    if args.mask_radius and not args.model:
        logger.write("You need model when you give mask_radius")
        return
    
    input_maps = []
    input_map_labels = []
    args.resolution -= 1e-6

    if args.map:
        logger.write("Input map: {}".format(args.map))
        map_obs = gemmi.read_ccp4_map(args.map).grid
        input_maps.append(map_obs)
        input_map_labels.append("obs")
        logger.write("Grid spacings: {} {} {}".format(*map_obs.spacing))
    else:
        map_obs = None
        
    if args.halfmaps:
        logger.write("Half map 1: {}".format(args.halfmaps[0]))
        logger.write("Half map 2: {}".format(args.halfmaps[1]))
        map_h1 = gemmi.read_ccp4_map(args.halfmaps[0]).grid
        map_h2 = gemmi.read_ccp4_map(args.halfmaps[1]).grid
        assert map_h1.shape == map_h2.shape
        assert map_h1.unit_cell == map_h2.unit_cell

        if map_obs is None:
            map_obs = utils.maps.half2full(map_h1, map_h2)
            input_maps.append(map_obs)
            input_map_labels.append("obs")
            logger.write("Grid spacings: {} {} {}".format(*map_obs.spacing))
        else:
            assert map_obs.shape == map_h1.shape
            assert map_obs.unit_cell == map_h1.unit_cell
            
        input_maps.extend([map_h1, map_h2])
        input_map_labels.extend(["half1", "half2"])
        
    unit_cell = map_obs.unit_cell
    spacegroup = gemmi.SpaceGroup(1)

    #if not args.map:
    #    map_obs = gemmi.FloatGrid((numpy.array(map_h1) + map_h2)/2, unit_cell, spacegroup)

    logger.write("Extent: {}".format(map_obs.shape))
    logger.write("Cell: {}".format(unit_cell.parameters))

    if args.mapref:
        if args.mapref == args.map:
            map_ref = map_obs
            logger.write("Using input map as a reference")
        else:
            map_ref = gemmi.read_ccp4_map(args.mapref).grid
            logger.write("Reference map: {}".format(args.mapref))
            assert unit_cell == map_ref.unit_cell
            assert map_obs.shape == map_ref.shape

        # Overwrite input_maps
        logger.write("Scaling maps..")
        input_maps = scale_maps(input_maps, map_ref, args.resolution)

    # Create mask
    mask = None
    if args.mask:
        logger.write("Input mask file: {}".format(args.mask))
        mask = numpy.array(gemmi.read_ccp4_map(args.mask).grid)
    
    st_new = None
    if args.model: # and 
        logger.write("Input model: {}".format(args.model))
        st = gemmi.read_structure(args.model)
        if args.remove_multiple_models and len(st) > 1:
            logger.write(" Removing models 2-{}".format(len(st)))
            for i in reversed(range(1, len(st))):
                del st[i]

        if args.ignore_symmetry and len(st.ncs) > 0:
            logger.write("Removing symmetry information from model.")
            st.ncs.clear()

        if args.relion_pg:
            logger.write("Point group symmetry given as RELION symbol: {}".format(args.relion_pg))
            if len(st.ncs) > 0:
                logger.write(" WARNING: NCS information in model file will be ignored")

            ops = utils.symmetry.get_matrices_using_relion(args.relion_pg)
            if ops:
                logger.write(" {} operators found".format(len(ops)))
                ops = utils.symmetry.make_NcsOps_from_matrices(ops, cell=unit_cell)
                st.ncs.clear()
                st.ncs.extend([x for x in ops if not x.tr.is_identity()])

        elif len(st.ncs) > 0:
            logger.write("Strict NCS detected from model.")

        if len(st.ncs) > 0:
            logger.write(" Writing NCS file")
            utils.symmetry.write_NcsOps_for_refmac(st.ncs, "ncsc_global.txt")
        
        st_new = st.clone()
        if len(st.ncs) > 0 and args.shift:
            logger.write("Expanding symmetry.")
            st.expand_ncs(gemmi.HowToNameCopiedChain.Short)
            logger.write(" Saving expanded model: input_model_expanded.*")
            utils.fileio.write_model(st, "input_model_expanded", pdb=True, cif=True)
    
        if not mask and args.mask_radius:
            logger.write("Creating mask..")
            mask = gemmi.FloatGrid(*map_obs.shape)
            mask.set_unit_cell(unit_cell)
            mask.spacegroup = spacegroup
            mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)
            ccp4 = gemmi.Ccp4Map()
            ccp4.grid = mask
            ccp4.update_ccp4_header(2, True) # float, update stats
            ccp4.write_ccp4_map("mask_from_model.ccp4")
            logger.write("Mask file written: mask_from_model.ccp4")

    # TODO Apply sharpening/blurring here?
    # 
    
    logger.write("Saving original maps as mtz files..")
    for ma, lab in zip(input_maps, input_map_labels):
        asu_obs = gemmi.transform_map_to_f_phi(ma).prepare_asu_data(dmin=args.resolution)
        write_map_mtz(asu_obs, args.output_mtz_prefix+"_"+lab+".mtz", blurs=args.blur)

    if mask:
        # Mask maps
        logger.write("Applying mask..")
        input_maps = [gemmi.FloatGrid(numpy.array(ma)*mask, unit_cell, spacegroup)
                      for ma in input_maps]

        if args.shift:
            logger.write(" Shifting maps and/or model..")
            tmp = numpy.where(numpy.array(mask)>0)
            limits = [(min(x), max(x)) for x in tmp]
            logger.write("  New limits: {} {} {}".format(*limits))
            min_points = [x[0] for x in limits]
            lu, lv, lw = limits
            # padding
            spacing = numpy.array(map_obs.spacing)
            p = args.mask_radius/spacing if args.mask_radius else 3/spacing
            p = p.astype(int)
            logger.write("  Spacing: {} {} {}".format(*map_obs.spacing))
            logger.write("  Padding: {} {} {}".format(*p))
            ll = [slice(max(0, limits[i][0]-p[i]),
                        min(limits[i][1]+p[i]+1, map_obs.shape[i])) for i in range(3)]
            #ll = [slice(28,120), slice(32,116), slice(36,112)]
            
            logger.write("  Slice {}".format(ll))
            
            # FIXME does not work for non-orthogonal cell
            new_shape = [ll[0].stop-ll[0].start,
                         ll[1].stop-ll[1].start,
                         ll[2].stop-ll[2].start]
            tmp = map_obs.point_to_position(*new_shape)
            new_cell = gemmi.UnitCell(tmp[0], tmp[1], tmp[2], 90, 90, 90)
            logger.write("  New Cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*new_cell.parameters))
            logger.write("  New grid: {} {} {}".format(*new_shape))

            if st_new:
                shifts = -mask.point_to_position(ll[0].start, ll[1].start, ll[2].start)
                logger.write("  Shifts: {:.4f} {:.4f} {:.4f}".format(*shifts.tolist()))
                for cra in st_new[0].all():
                    cra.atom.pos += shifts
                
                st_new.cell = new_cell
                st_new.spacegroup_hm = "P 1"
                if len(st_new.ncs) > 0:
                    new_ops = utils.symmetry.apply_shift_for_ncsops(st_new.ncs, shifts)
                    st_new.ncs.clear()
                    st_new.ncs.extend(new_ops)
                    logger.write(" Writing NCS file for shifted model")
                    utils.symmetry.write_NcsOps_for_refmac(st_new.ncs, "ncsc_local.txt")
                    logger.write(" Writing symmetry expanded model for shifted model")
                    utils.symmetry.write_symmetry_expanded_model(st_new, "shifted_local_expanded",
                                                                 pdb=True, cif=True)

                logger.write(" Saving shifted model..")
                utils.fileio.write_model(st_new, "shifted_local", pdb=True, cif=True)


            logger.write(" Saving masked and shifted maps as mtz files..")
            for ma, lab in zip(input_maps, input_map_labels):
                logger.write("  Processing {} map".format(lab))
                org_grid = numpy.array(ma)
                new_grid = org_grid[ll[0], ll[1], ll[2]]

                new_grid = gemmi.FloatGrid(new_grid, new_cell, spacegroup)
                asu_new = gemmi.transform_map_to_f_phi(new_grid).prepare_asu_data(dmin=args.resolution)
                write_map_mtz(asu_new, args.output_mtz_prefix+"_"+lab+".mtz", blurs=args.blur)

            json.dump(dict(cell=unit_cell.parameters,
                           grid=map_obs.shape,
                           new_cell=new_cell.parameters,
                           new_grid=list(map(int, new_shape)),
                           shifts=shifts.tolist()),
                      open("shifts.json", "w"), indent=2)
        else:
            logger.write(" Saving input model with unit cell information")
            st.cell = unit_cell
            st.spacegroup_hm = "P 1"
            utils.fileio.write_model(st, "starting_model", pdb=True, cif=True)
            
            logger.write(" Saving masked maps as mtz files..")
            for ma, lab in zip(input_maps, input_map_labels):
                logger.write("  Processing {} map".format(lab))
                asu_new = gemmi.transform_map_to_f_phi(ma).prepare_asu_data(dmin=args.resolution)
                write_map_mtz(asu_new, args.output_mtz_prefix+"_"+lab+".mtz", blurs=args.blur)

    # TODO write NCS file for REFMAC5
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
import gemmi
import numpy
import os

def load_monomer_library(resnames, monomer_dir=None): #TODO add cif_paths
    if monomer_dir is None:
        if "CLIBD_MON" not in os.environ:
            logger.write("ERROR: CLIBD_MON is not set")
            return
        monomer_dir = os.environ["CLIBD_MON"]
        
    if not os.path.isdir(monomer_dir):
        logger.write("ERROR: not a directory: {}".format(monomer_dir))
        return
        
    monlib = gemmi.read_monomer_lib(monomer_dir, resnames)
    return monlib
# load_monomer_library()

def shake_structure(st, sigma):
    print("Randomizing structure with rmsd of {}".format(sigma))
    st2 = st.clone()
    for model in st2:
        for cra in model.all():
            r = numpy.random.normal(0, sigma, 3)
            cra.atom.pos += gemmi.Position(*r)

    return st2
# shake_structure()

def check_monlib_support_nucleus_distances(monlib, resnames):
    good = True
    for resn in resnames:
        if resn not in monlib.monomers:
            logger.write("ERROR: monomer information of {} not loaded".format(resn))
            good = False
        else:
            mon = monlib.monomers[resn]
            no_nuc = False
            for bond in mon.rt.bonds:
                is_h = (mon.get_atom(bond.id1.atom).is_hydrogen(), mon.get_atom(bond.id2.atom).is_hydrogen())
                if any(is_h) and bond.value_nucleus != bond.value_nucleus:
                    no_nuc = True
                    break
            if no_nuc:
                logger.write("ERROR: nucleus distance is not found for {}".format(resn))
                good = False

    return good

def calc_fc_em(st, d_min, mott_bethe=True, monlib=None, blur=0, r_cut=1e-5, rate=1.5):
    # XXX check minimum B and set appropriate blur

    if monlib is not None and st[0].count_hydrogen_sites() > 0:
        st = st.clone()
        topo = gemmi.prepare_topology(st, monlib)
        resnames = st[0].get_all_residue_names()
        check_monlib_support_nucleus_distances(monlib, resnames)
    else:
        topo = None
        
    if mott_bethe:
        dc = gemmi.DensityCalculatorX()
        dc.d_min = d_min
        dc.blur = blur
        dc.r_cut = r_cut
        dc.rate = rate
        dc.set_grid_cell_and_spacegroup(st)

        if topo is None:
            dc.addends.subtract_z()
            dc.put_model_density_on_grid(st[0])
        else:
            # 
            dc.initialize_grid()
            dc.addends.subtract_z(except_hydrogen=True)
            dc.add_model_density_to_grid(st[0])
            # 
            topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus)
            for cra in st[0].all():
                if cra.atom.is_hydrogen():
                    dc.add_c_contribution_to_grid(cra.atom, -1)
                    
            dc.sum_symmetry_equivalent_grid_points()

        grid = gemmi.transform_map_to_f_phi(dc.grid)
        asu_data = grid.prepare_asu_data(dmin=d_min, mott_bethe=True, unblur=dc.blur)
    else:
        dc = gemmi.DensityCalculatorE()
        dc.d_min = d_min
        dc.blur = blur
        d.r_cut = r_cut
        dc.rate = rate
        dc.addends.subtract_z()
        dc.set_grid_cell_and_spacegroup(st)
        dc.put_model_density_on_grid(st[0])
        grid = gemmi.transform_map_to_f_phi(dc.grid)
        asu_data = grid.prepare_asu_data(dmin=d_min)
        
    return asu_data

# calc_fc_em()

def all_B(st):
    ret = []
    for mol in st:
        for chain in mol: # TODO use all()!
            for res in chain:
                for at in res:
                    ret.append(at.b_iso)

    return ret
# all_B()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import sys
import datetime
import pipes
import servalcat.spa.sfcalc
import servalcat.spa.shiftback
import servalcat.spa.run_refmac
import servalcat.spa.fsc
import servalcat.spa.fofc
import servalcat.spa.shift_maps
#import servalcat.spa.bestmap

from servalcat.utils import logger

def main():
    logger.set_file("servalcat.log")
    logger.write("# Started on {}".format(datetime.datetime.now()))
    logger.write("# Command-line args:")
    logger.write("# {}".format(" ".join(map(lambda x: pipes.quote(x), sys.argv[1:]))))

    
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command", description='')

    sfcalc = subparsers.add_parser('sfcalc')
    servalcat.spa.sfcalc.add_arguments(sfcalc)
    shiftback = subparsers.add_parser('shiftback')
    servalcat.spa.shiftback.add_arguments(shiftback)
    refine_spa = subparsers.add_parser('refine_spa')
    servalcat.spa.run_refmac.add_arguments(refine_spa)
    fsc = subparsers.add_parser('fsc')
    servalcat.spa.fsc.add_arguments(fsc)
    fofc = subparsers.add_parser('fofc')
    servalcat.spa.fofc.add_arguments(fofc)
    shift_maps = subparsers.add_parser('shift')
    servalcat.spa.shift_maps.add_arguments(shift_maps)
    #bestmap = subparsers.add_parser('bestmap')
    #servalcat.spa.bestmap.add_arguments(bestmap)

    args = parser.parse_args()

    if args.command == "sfcalc":
        servalcat.spa.sfcalc.main(args)
    elif args.command == "shiftback":
        servalcat.spa.shiftback.main(args)
    elif args.command == "refine_spa":
        servalcat.spa.run_refmac.main(args)
    elif args.command == "fsc":
        servalcat.spa.fsc.main(args)
    elif args.command == "fofc":
        servalcat.spa.fofc.main(args)
    elif args.command == "shift":
        servalcat.spa.shift_maps.main(args)
    #elif args.command == "bestmap":
    #    servalcat.spa.bestmap.main(args)
    else:
        parser.print_help()
        
    logger.write("# Finished on {}".format(datetime.datetime.now()))
# main()

if __name__ == "__main__":
    main()




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import os
import subprocess
import gemmi
import numpy
import argparse

def splitext(path):
    if path.endswith((".bz2",".gz")):
        return os.path.splitext(path[:path.rindex(".")])
    else:
        return os.path.splitext(path)
# splitext()

def check_model_format(xyzin):
    # TODO check format actually
    # TODO mmjson is possible?
    ext = splitext(xyzin)[1]
    if ext.endswith("cif"):
        return ".mmcif"
    else:
        return ".pdb"
# check_model_format()

def write_mmcif(st, cif_out, cif_ref=None):
    """
    Refmac fails if _entry.id is longer than 80 chars including quotations
    """
    st_new = st.clone()
    print("Writing mmCIF file:", cif_out)
    if cif_ref:
        print("  using mmCIF metadata from:", cif_ref)
        groups = gemmi.MmcifOutputGroups(False)
        groups.ncs = True
        groups.atoms = True
        groups.cell = True
        doc = gemmi.cif.read(cif_ref)
        block = doc.find_block(st_new.info["_entry.id"])
        st_new.update_mmcif_block(block, groups)
        st_new.info["_entry.id"] = st_new.info["_entry.id"][:78]
        doc.write_file(cif_out)
    else:
        st_new.name = st_new.name[:78] # this will become _entry.id
        st_new.make_mmcif_document().write_file(cif_out)
# write_mmcif()

def write_pdb(st, pdb_out):
    print("Writing shifted PDB:", pdb_out)
    st_new = st.clone()
    st_new.shorten_chain_names()
    st_new.write_pdb(pdb_out, use_linkr=True)
# write_pdb()

def write_model(st, prefix=None, file_name=None, pdb=False, cif=False, cif_ref=None):
    if file_name:
        if file_name.endswith("cif"):
            write_mmcif(st, file_name, cif_ref)
        elif file_name.endswith(".pdb"):
            write_pdb(st, file_name)
        else:
            raise Exception("Cannot determine file format from file name: {}".format(file_name))
    else:            
        if cif:
            write_mmcif(st, prefix+".mmcif", cif_ref)
        if pdb:
            write_pdb(st, prefix+".pdb")
# write_model()

def read_shifts_txt(shifts_txt):
    ret = {}
    s = open(shifts_txt).read()
    s = s.replace("\n"," ").split()
    for i in range(len(s)-3):
        if s[i] in ("pdbin", "pdbout") and s[i+1] in ("cell", "shifts"):
            n = 6 if s[i+1] == "cell" else 3
            ret["{} {}".format(s[i], s[i+1])] = [float(x) for x in s[i+2:i+2+n]]

    return ret
# read_shifts_txt()

def read_map_from_mtz(mtz_in, cols, grid_size=None, sample_rate=3):
    mtz = gemmi.read_mtz_file(mtz_in)
    d_min = mtz.resolution_high() # TODO get resolution for column?
    if grid_size is None:
        grid_size = mtz.get_size_for_hkl(sample_rate=sample_rate)
    F = mtz.get_f_phi_on_grid(cols[0], cols[1], grid_size)
    m = gemmi.transform_f_phi_grid_to_map(F)
    return d_min, m
# read_map_from_mtz()

def read_asu_data_from_mtz(mtz_in, cols):
    assert 0 < len(cols) < 3
    mtz = gemmi.read_mtz_file(mtz_in)
    sg = mtz.spacegroup
    miller = mtz.make_miller_array()
    f = mtz.column_with_label(cols[0])
    cell = mtz.get_cell(f.dataset_id)
    if len(cols) == 2:
        phi = mtz.column_with_label(cols[1])
        assert f.type == "F"
        assert phi.type == "P"
        phi = numpy.deg2rad(phi)
        f_comp = f * (numpy.cos(phi) + 1j * numpy.sin(phi))
        asu = gemmi.ComplexAsuData(cell, sg, miller, f_comp)
        return asu
    else:
        if f.is_integer():
            gr_t = gemmi.Int8AsuData # Why Int8?? XXX
        else:
            gr_t = gemmi.FloatAsuData
        
        asu = gr_t(cell, sg, miller, f)
        return asu
# read_asu_data_from_mtz()


def read_structure_from_pdb_and_mmcif(xyz_in):
    st = gemmi.read_structure(xyz_in)
    cif_ref = None
    if xyz_in.endswith(".pdb"):
        cif_in = xyz_in[:-4] + ".mmcif"
        if os.path.isfile(cif_in):
            print(" Will use mmcif metadata from {}".format(cif_in))
            cif_ref = cif_in
    elif xyz_in.endswith(".mmcif"):
        cif_ref = xyz_in
        pdb_in = xyz_in[:-6] + ".pdb"
        if os.path.isfile(pdb_in):
            print(" Reading PDB REMARKS from {}".format(pdb_in))
            tmp = gemmi.read_structure(pdb_in)
            st.raw_remarks = tmp.raw_remarks

    return st, cif_ref
# read_structure_from_pdb_and_mmcif()

def merge_ligand_cif(cifs_in, cif_out):
    docs = [gemmi.cif.read(x) for x in cifs_in]
    tags = ["_chem_comp.id"]
    for d in docs:
        b = d.find_block("comp_list")
        l = b.find_loop("_chem_comp.id").get_loop()
        for t in l.tags:
            if t not in tags: tags.append(t)
  
    doc = gemmi.cif.Document()
    comp_list = doc.add_new_block("comp_list")
    loop = comp_list.init_loop("", tags)
    tags_for_find = [tags[0]] + ["?"+x for x in tags[1:]]
    for d in docs:
        b = d.find_block("comp_list")
        vals = b.find(tags_for_find)
        for v in vals:
            rl = [v.get(x) if v.has(x) else "." for x in range(len(tags))]
            loop.add_row(rl)

    for d in docs:
        for b in d:
            if b.name != "comp_list":
                doc.add_copied_block(b)

    doc.write_file(cif_out)
# merge_ligand_cif()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import numpy
import pandas
import gemmi

class Binner:
    def __init__(self, asu, style="relion"):
        if style == "relion":
            cell = asu.unit_cell
            max_cell_edge = numpy.max([cell.a, cell.b, cell.c])
            self.d_array = asu.make_d_array()
            self.bin_array = (max_cell_edge/self.d_array).astype(int)
            self.bins, self.bin_counts = numpy.unique(self.bin_array, return_counts=True)
        else:
            raise Exception("Non-supported binning type")
    # __init__()
# class Binner    

def df_from_asu_data(asu_data, label):
    df = pandas.DataFrame(data=asu_data.miller_array,
                          columns=["H","K","L"])
    df[label] = asu_data.value_array
    return df

def df_from_raw(miller_array, value_array, label):
    df = pandas.DataFrame(data=miller_array,
                          columns=["H","K","L"])
    df[label] = value_array
    return df

class HklData:
    def __init__(self, cell, sg, anomalous, df, binned_df=None):
        self.cell = cell
        self.sg = sg
        self.anomalous = anomalous
        self.df = df
        self.binned_df = binned_df
    # __init__()

    def update_cell(self, cell):
        # update d
        pass

    def switch_to_asu(self):
        # Need to care phases
        pass

    def copy(self):
        return HklData(self.cell, self.sg, self.anomalous,
                       self.df.copy(),
                       self.binned_df.copy() if self.binned_df is not None else None)

    def merge_asu_data(self, asu_data, label, common_only=True):
        if label in self.df:
            raise Exception("Duplicated label")
        
        df_tmp = df_from_asu_data(asu_data, label)
        if common_only:
            self.df = self.df.merge(df_tmp)
        else:
            self.df = self.df.merge(other.df, how="outer")
    # merge_asu_data()

    def miller_array(self):
        return self.df[["H","K","L"]]

    def calc_d(self):
        self.df["d"] = self.cell.calculate_d_array(self.miller_array())
    # calc_d()

    def d_spacings(self):
        if "d" not in self.df or self.df.d.isnull().values.any():
            self.calc_d()
        return self.df.d
    # calc_d()

    def d_min_max(self):
        d = self.d_spacings()
        return min(d), max(d)
    # d_min_max()

    def setup_relion_binning(self):
        max_edge = max(self.cell.parameters[:3])
        if "d" not in self.df or self.df.d.isnull().values.any():
            self.calc_d()
            
        self.df["bin"] = (max_edge/self.df.d).astype(numpy.int)
        self.binned_df = pandas.DataFrame(data=list(range(max(self.df.bin)+1)),
                                          columns=["bin"])
    # setup_relion_binning()

    def bin_and_limits(self):
        bins = set(self.df.bin)
        ret = []
        for i_bin in bins:
            sel = self.df.bin == i_bin
            d_sel = self.df.d[sel]
            ret.append((i_bin, max(d_sel), min(d_sel)))

        return ret
    # bin_and_limits()
    
    def merge(self, other, common_only=True):
        self.merge_df(other, common_only)
    # merge()

    def merge_df(self, other, common_only=True):
        # TODO check space group, cell, anomalous
        # TODO transform to asu with phase shifts
        # TODO check column labels. same names other than HKL?
        # 
        if common_only:
            self.df = self.df.merge(other)
        else:
            df = self.df.merge(other, indicator=True, how="outer")
            df_left = df[df._merge=="left_only"]
            df_right = df[df._merge=="right_only"]
            df_both = df[df._merge=="both"]
    # merge()

    def fft_map(self, label, grid_size=None, sample_rate=3):
        asu = gemmi.ComplexAsuData(self.cell,
                                   self.sg,
                                   self.miller_array(),
                                   self.df[label])
        if grid_size is None:
            grid_size = (0, 0, 0)

        ma = asu.transform_f_phi_to_map(sample_rate=sample_rate, exact_size=grid_size)
        return ma
    # fft_map()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import os
import subprocess
import pipes
from servalcat.utils import logger
from servalcat import utils
from servalcat import spa

def add_arguments(parser):
    parser.description = 'Run REFMAC5 for SPA'

    parser.add_argument('--exe', default="refmac5", help='refmac5 binary')
    # sfcalc options
    parser.add_argument('--map', help='Input map file')
    parser.add_argument('--halfmaps', nargs=2, help='Input half map files')
    parser.add_argument('--mapref', help='Reference map file')
    parser.add_argument('--mask', help='Mask file')
    parser.add_argument('--model', required=True, help="")
    parser.add_argument('--mask_radius', type=float, help='')
    parser.add_argument('--resolution', type=float, help='')
    parser.add_argument('--shift', action='store_true', help='')
    parser.add_argument('--blur', nargs="+", type=float, help='Sharpening or blurring B')
    parser.add_argument('--ligand', nargs="*")
    parser.add_argument('--relion_pg',
                        help='RELION point group symbol for strict symmetry')
    parser.add_argument('--ignore_symmetry', action='store_true',
                        help='Ignore symmetry information in the model file')
    parser.add_argument('--remove_multiple_models', action='store_true',
                        help='Keep 1st model only')
    # run_refmac options
    parser.add_argument('--mtz', help='Input mtz file')
    parser.add_argument('--mtz_half', nargs=2, help='Input mtz files for half maps')
    parser.add_argument('--lab_f')
    parser.add_argument('--lab_sigf')
    parser.add_argument('--lab_phi')
    parser.add_argument('--bfactor', type=float)
    parser.add_argument('--ncsr', default="local")
    parser.add_argument('--ncycle', type=int, default=10)
    parser.add_argument('--hydrogen', default="all")
    parser.add_argument('--jellybody', action='store_true')
    parser.add_argument('--hout', action='store_true')
    parser.add_argument('--weight_auto_scale', type=float)
    parser.add_argument('--keywords', nargs='+')
    parser.add_argument('--keyword_file', nargs='+')
    parser.add_argument('--output_prefix', default="refined",
                        help='output file name prefix')
    parser.add_argument('--cross_validation', action='store_true',
                        help='Run cross validation')
    parser.add_argument('--shake_radius', default=0.3,
                        help='Shake rmsd')

# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def make_cmd(xyzin, hklin, xyzout=None, hklout=None, prefix=None, exe="refmac5", libin=None):
    cmd = [exe]
    if prefix:
        if not xyzout: xyzout = prefix + ".pdb"
        if not hklout: hklout = prefix + ".mtz"
        
    cmd.extend(["hklin", pipes.quote(hklin)])
    cmd.extend(["hklout", pipes.quote(hklout)])
    cmd.extend(["xyzin", pipes.quote(xyzin)])
    cmd.extend(["xyzout", pipes.quote(xyzout)])
    if libin:
        if len(libin) > 1:
            mcif = "merged_ligands.cif"
            logger.write("Merging ligand cif inputs: {}".format(libin))
            utils.fileio.merge_ligand_cif(libin, mcif)
            cmd.extend(["libin", mcif])
        else:
            cmd.extend(["libin", libin[0]])

    return cmd
# make_cmd()

def make_keywords(lab_f, lab_sigf, lab_phi, ncycle, hydrogen="all", hout=False, bfactor=None, jellybody=True,
                  resolution=None, weight_fixed=None, weight_auto_scale=None, ncsr=None,
                  keyword_files=None, keywords=None):
    ret = ""
    labin = []
    if lab_f: labin.append("FP={}".format(lab_f))
    if lab_sigf: labin.append("SIGFP={}".format(lab_sigf))
    if lab_phi: labin.append("PHIB={}".format(lab_phi))
    ret += "labin {}\n".format(" ".join(labin))

    
    ret += "make hydr {}\n".format(hydrogen)
    ret += "make hout {}\n".format("yes" if hout else "no")
    ret += "solvent no\n"
    ret += "source em mb\n"
    ret += "ncycle {}\n".format(ncycle)
    if resolution is not None:
        ret += "reso {}\n".format(resolution)
    if weight_fixed is not None:
        ret += "weight matrix {}\n".format(weight_fixed)
    elif weight_auto_scale is not None:
        ret += "weight auto {}\n".format(weight_auto_scale)
    else:
        ret += "weight auto\n"

    if bfactor is not None:
        ret += "bfactor set {}\n".format(bfactor)
    if jellybody:
        ret += "ridge dist sigma 0.01\n"
        ret += "ridge dist dmax 4.2\n"
    if ncsr:
        ret += "ncsr {}\n".format(ncsr)
    
    if keyword_files:
        for f in keyword_files:
            ret += "@{}\n".format(f)

    if keywords:
        ret += keywords

    return ret
# make_keywords()

def run_refmac(mtz_in, model_in, ncycle, lab_f, lab_phi, lab_sigf=None, hydrogen="auto", hout=False,
               libin=None, bfactor=None, jellybody=None, resolution=None, weight_auto_scale=None,
               ncsr=None, keyword_files=None, keywords=None,
               prefix="refined", exe="refmac5"):
    cmd = make_cmd(model_in, mtz_in, libin=libin, prefix=prefix, exe=exe)
    
    logger.write("Running REFMAC5..")
    log = open(prefix+".log", "w")
    p = subprocess.Popen(cmd, shell=False, stdin=subprocess.PIPE, stdout=log, stderr=log)
    stdin = make_keywords(lab_f=lab_f, lab_phi=lab_phi, lab_sigf=lab_sigf,
                          ncycle=ncycle, resolution=resolution,
                          hydrogen=hydrogen, hout=hout, ncsr=ncsr,
                          bfactor=bfactor, jellybody=jellybody,
                          weight_auto_scale=weight_auto_scale,
                          keyword_files=keyword_files,
                          keywords=keywords)
    open(prefix+".inp", "w").write(stdin)
    p.stdin.write(stdin.encode("utf-8"))
    p.stdin.close()
    ret = p.wait()
    logger.write("REFMAC5 finished with {}".format(ret))

# run_refmac()

def main(args):
    model_format = utils.fileio.check_model_format(args.model)
    if args.map or args.halfmaps:
        args.output_model_prefix = "shifted_local"
        args.output_mtz_prefix = "masked_fs"
        args.remove_multiple_models = True
        spa.sfcalc.main(args)
        args.mtz = "masked_fs_obs.mtz"
        if args.halfmaps:
            args.mtz_half = ["masked_fs_half1.mtz", "masked_fs_half2.mtz"]
        args.lab_phi = "Pout0"
        if args.blur:
            args.lab_f = "FoutBlur_{:.2f}".format(args.blur[0])
        else:
            args.lab_f = "Fout0"
        if args.shift:
            args.model = "shifted_local" + model_format
        else:
            args.model = "starting_model" + model_format

    keyword_files = []
    if args.keyword_file:
        for f in args.keyword_file:
            logger.write("Keyword file: {}".format(f))
            assert os.path.exists(f)
            keyword_files.append(f)
        
    keywords = ""
    if args.keywords:
        keywords = "\n".join(args.keywords)
            
    if args.shift:
        refmac_prefix = args.output_prefix + "_local"
        if os.path.isfile("ncsc_local.txt"):
            keyword_files.append("ncsc_local.txt")
    else:
        refmac_prefix = args.output_prefix
        if os.path.isfile("ncsc_global.txt"):
            keyword_files.append("ncsc_global.txt")


    run_refmac(mtz_in=args.mtz,
               model_in=args.model,
               ncycle=args.ncycle,
               lab_f=args.lab_f, lab_phi=args.lab_phi, lab_sigf=args.lab_sigf,
               hydrogen=args.hydrogen, hout=args.hout,
               ncsr=args.ncsr,
               libin=args.ligand,
               bfactor=args.bfactor,
               jellybody=args.jellybody,
               resolution=args.resolution,
               prefix=refmac_prefix,
               weight_auto_scale=args.weight_auto_scale,
               keyword_files=keyword_files,
               keywords=keywords,
               exe=args.exe)

    """
    # Calc diffmaps
    if 0:
        st = gemmi.read_structure(refmac_prefix+model_format)
        st.expand_ncs(gemmi.HowToNameCopiedChain.Short)
        grid_shape = 
        mask = gemmi.FloatGrid(*grid_shape)
        mask.set_unit_cell(st.cell)
        mask.spacegroup = gemmi.SpaceGroup(st.spacegroup_hm)
        mask.mask_points_in_constant_radius(st[0], args.mask_radius, 1.)
        spa.fofc.
    """

    if args.shift:
        ncsc_in = ("ncsc_global.txt") if os.path.isfile("ncsc_global.txt") else None
        spa.shiftback.shift_back(xyz_in=refmac_prefix+model_format,
                                 refine_mtz=refmac_prefix+".mtz",
                                 shifts_json="shifts.json",
                                 ncsc_in=ncsc_in,
                                 out_prefix=args.output_prefix)

    # Expand sym here
    if os.path.isfile("ncsc_global.txt"):
        refined_xyz = args.output_prefix+model_format
        logger.write("Expanding {}".format(refined_xyz))
        st, cif_ref = utils.fileio.read_structure_from_pdb_and_mmcif(refined_xyz)
        st.expand_ncs(gemmi.HowToNameCopiedChain.Short)
        utils.fileio.write_model(st, file_name=args.output_prefix+"_expanded"+model_format,
                                 cif_ref=cif_ref)


    if args.cross_validation:
        st = gemmi.read_structure(refmac_prefix+model_format)
        st = utils.model.shake_structure(st, args.shake_radius)
        shaken_file = refmac_prefix+"_shaken"+model_format
        utils.fileio.write_model(st, file_name=shaken_file)
        if args.shift:
            refmac_prefix_shaken = refmac_prefix+"_shaken_refined_local"
        else:
            refmac_prefix_shaken = refmac_prefix+"_shaken_refined"

        run_refmac(mtz_in=args.mtz_half[0],
                   model_in=shaken_file,
                   ncycle=args.ncycle,
                   lab_f=args.lab_f, lab_phi=args.lab_phi, lab_sigf=args.lab_sigf,
                   hydrogen=args.hydrogen, hout=args.hout,
                   bfactor=args.bfactor,
                   jellybody=args.jellybody,
                   resolution=args.resolution,
                   prefix=refmac_prefix_shaken,
                   keyword_files=keyword_files,
                   keywords=keywords)

        # TODO calc FSC

        if args.shift:
            ncsc_in = ("ncsc_global.txt") if os.path.isfile("ncsc_global.txt") else None
            spa.shiftback.shift_back(xyz_in=refmac_prefix_shaken+model_format,
                                 refine_mtz=refmac_prefix_shaken+".mtz",
                                 shifts_json="shifts.json",
                                 ncsc_in=ncsc_in,
                                 out_prefix=refmac_prefix+"_shaken_refined")
        
        
if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)

